/* crc32.c -- compute the CRC-32 of a data stream
 * Copyright (C) 1995-2002 Mark Adler
 * For conditions of distribution and use, see copyright notice in zlib.h
 */

/* @(#) $Id: crc32.c,v 1.4 2006/04/15 11:39:24 arpruss Exp $ */

#ifdef ARMLET
#include <Standalone.h>
#include <PalmOS.h>
#include <PceNativeCall.h>
#include "crc32.h"
#include "Backup_Res.h"
#include "endianutils.h"
#define PtrByteSwap32( x ) ( ( void* )ByteSwap32( ( UInt32 )( x ) ) )
STANDALONE_CODE_RESOURCE_ID( armCRC );

UInt32 do_crc32(UInt32* crc_table, UInt32 crc, UInt8* buf, UInt32 len);

UInt32 NativeFunction
    (
    const void*      emulStateP,
    void*            userData68KP,
    Call68KFuncType* call68KFuncP
    )
{
    CRCDataType* c = userData68KP;

    return do_crc32( c->table, ByteSwap32( c->crc ), PtrByteSwap32( c->buf ), ByteSwap32( c->length ) );
}
#else
#include <PalmOS.h>
#include <PceNativeCall.h>
#include "Backup.h"
#include "crc32.h"

UInt32 do_crc32(UInt32* crc_table, UInt32 crc, UInt8* buf, UInt32 len);

static MemHandle armH = NULL;
static void*     ArmCRCFunction;
static CRCDataType* crcData;

void InitCRC( void )
{
    UInt32 processor;

    CloseCRC();

    crcData = SafeMemPtrNew( sizeof( CRCDataType ) );

    MemSet( crcData, sizeof( *crcData ), 0 );

    if ( errNone == FtrGet(sysFileCSystem, sysFtrNumProcessorID, &processor ) &&
         sysFtrNumProcessorIsARM( processor ) &&
         NULL != ( armH = DmGetResource( 'armc', armCRC ) )
          ) {
        ArmCRCFunction = MemHandleLock( armH );
    }
    else {
        armH = NULL;
    }
}



void CloseCRC( void )
{
    if ( armH != NULL ) {
        MemHandleUnlock( armH );
        DmReleaseResource( armH );
        ArmCRCFunction = NULL;
        armH = NULL;
    }

    SafeMemPtrFree( crcData );
    crcData = NULL;
}




UInt32 CRC32( UInt32 crc, UInt8* buf, UInt32 len )
{
    if ( armH != NULL ) {
        crcData->crc   = crc;
        crcData->buf   = buf;
        crcData->length = len;

        crc = PceNativeCall( ArmCRCFunction, crcData );
    }
    else {
        crc = do_crc32( crcData->table, crc, buf, len );
    }

    return crc;
}
#endif


void make_crc_table( UInt32* table );

/* ========================================================================= */
#define DO1(buf) crc = crc_table[((int)crc ^ (*buf++)) & 0xff] ^ (crc >> 8);
#define DO2(buf)  DO1(buf); DO1(buf);
#define DO4(buf)  DO2(buf); DO2(buf);
#define DO8(buf)  DO4(buf); DO4(buf);

/* ========================================================================= */
UInt32 do_crc32(UInt32* crc_table, UInt32 crc, UInt8* buf, UInt32 len)
{
    if ( crc_table[ 1 ] != 0x77073096L ) {
        make_crc_table( crc_table );
    }
    crc = crc ^ 0xffffffffL;
    while (len >= 8)
    {
      DO8(buf);
      len -= 8;
    }
    if (len) do {
      DO1(buf);
    } while (--len);
    return crc ^ 0xffffffffL;
}


/*
  Generate a table for a byte-wise 32-bit CRC calculation on the polynomial:
  x^32+x^26+x^23+x^22+x^16+x^12+x^11+x^10+x^8+x^7+x^5+x^4+x^2+x+1.

  Polynomials over GF(2) are represented in binary, one bit per coefficient,
  with the lowest powers in the most significant bit.  Then adding polynomials
  is just exclusive-or, and multiplying a polynomial by x is a right shift by
  one.  If we call the above polynomial p, and represent a byte as the
  polynomial q, also with the lowest power in the most significant bit (so the
  byte 0xb1 is the polynomial x^7+x^3+x+1), then the CRC is (q*x^32) mod p,
  where a mod b means the remainder after dividing a by b.

  This calculation is done using the shift-register method of multiplying and
  taking the remainder.  The register is initialized to zero, and for each
  incoming bit, x^32 is added mod p to the register if the bit is a one (where
  x^32 mod p is p+x^32 = x^26+...+1), and the register is multiplied mod p by
  x (which is shifting right by one and adding x^32 mod p if the bit shifted
  out is a one).  We start with the highest power (least significant bit) of
  q and repeat for all eight bits of q.

  The table is simply the CRC of all possible eight bit values.  This is all
  the information needed to generate CRC's on data a byte at a time for all
  combinations of CRC register values and incoming bytes.
*/
void make_crc_table( UInt32* crc_table )
{
  UInt32 c;
  UInt32 n, k;
  UInt32 poly;            /* polynomial exclusive-or pattern */
  UInt8  p[14];
  /* terms of polynomial defining this crc (except x^32): */
  p[0]=0;
  p[1]=1;
  p[2]=2;
  p[3]=4;
  p[4]=5;
  p[5]=7;
  p[6]=8;
  p[7]=10;
  p[8]=11;
  p[9]=12;
  p[10]=16;
  p[11]=22;
  p[12]=23;
  p[13]=26;

  /* make exclusive-or pattern from polynomial (0xedb88320L) */
  poly = 0L;
  for (n = 0; n < sizeof(p)/sizeof(UInt8); n++)
    poly |= 1L << (31 - p[n]);

  for (n = 0; n < 256; n++)
  {
    c = (UInt32)n;
    for (k = 0; k < 8; k++)
      c = c & 1 ? poly ^ (c >> 1) : c >> 1;
    crc_table[n] = c;
  }
}

