/* $Id: ftp.c,v 1.12 2007/03/12 02:19:47 arpruss Exp $ */
#include <PalmOS.h>
#include <NetMgr.h>
#include <VFSMgr.h>
#include <Unix/sys_socket.h>
#include "Backup.h"
#include "ftp.h"

#define RESPONSE_TIMEOUT     60//180
#define CONNECT_TIMEOUT      60
#define SEND_COMMAND_TIMEOUT 60

#define FTP_COMMAND_LENGTH 512
#define MISC_BUFFER_SIZE   80 // minimum 5
#define RESPONSE_START_SIZE 300
#define DATA_BUFFER_SIZE    4096

#define MY_ERROR 9999

#ifndef SHUT_RDWR
# define SHUT_RDWR netSocketDirBoth
#endif

typedef struct _DirListNode {
    Char* name;
    struct _DirListNode* next;
} DirListNode;

static Char*        curName = NULL;
static DirListNode* dirListStart = NULL;
static DirListNode* dirListCur   = NULL;
static UInt16 curNamePos = 0;
static Char*  writeBuffer;
static UInt32  writeBufferSize;

typedef Err ( *ReaderFunction )( UInt32 ref, UInt32 size, void* buf, UInt32* didReadP );
typedef Err ( *WriterFunction )( UInt32 ref, UInt32 size, const void* buf, UInt32* didReadP );

UInt16 errno;

static NetSocketRef controlSocket = -1;
static NetSocketRef dataSocket    = -1;

static Char command[ FTP_COMMAND_LENGTH ];
static Char responseStart[ RESPONSE_START_SIZE ];
static Char miscBuffer[ MISC_BUFFER_SIZE ];
static NetSocketAddrINType passiveAddress;
static Boolean             havePassive = false;
Boolean             netOpen = false;

static Int32 oldTimeout = -1;

FTP_SECTION static Err BufferWrite( UInt32 ref, UInt32 size, const void* buf, UInt32* didWriteP );
FTP_SECTION static Err DirSave( UInt32 ref, UInt32 toSave, const void* buffer, UInt32* didSaveP );
FTP_SECTION static void ClosePassiveDataSocket( void );
FTP_SECTION static Boolean OpenPassiveDataSocket( void );
FTP_SECTION static Err MyVFSFileRead( UInt32 ref, UInt32 size, void* buf, UInt32* didReadP );
FTP_SECTION static Err MyVFSFileWrite( UInt32 ref, UInt32 size, const void* buf, UInt32* didReadP );
FTP_SECTION static Boolean FTPSend( const Char* dest,
     ReaderFunction Reader, UInt32 ref, UInt32 size );
FTP_SECTION static Boolean FTPFetchData( Char* fetchCommand,
    WriterFunction Writer, UInt32 ref );
FTP_SECTION static Err DirSave( UInt32 ref, UInt32 toSave, const void* buffer, UInt32* didSaveP );
FTP_SECTION static void NetCloseGeneric( Boolean immediate );

void FTPClose( void )
{
    if ( 0 <= dataSocket ) {
        shutdown( dataSocket, SHUT_RDWR );
        close( dataSocket );
        dataSocket = -1;
    }

    if ( 0 <= controlSocket ) {
        shutdown( controlSocket, SHUT_RDWR );
        close( controlSocket );
        controlSocket = -1;
    }

    havePassive = false;
}


static void NetCloseGeneric( Boolean immediate )
{
    if ( 0 <= oldTimeout ) {
        AppNetTimeout = oldTimeout;
        oldTimeout = -1;
    }

//    SysTaskDelay( SysTicksPerSecond() * 2 );

    if ( netOpen ) {
        NetLibClose( AppNetRefnum, false ); //non-immediate close
        netOpen = false;
    }
}



void NetClose( void )
{
    NetCloseGeneric( false );
}



void NetCloseImmediate( void )
{
    NetCloseGeneric( true );
}


UInt16 FTPResponse( void )
{
    Char    code[ 4 ];
    Boolean lastLine;
    Int16   scannedNumber = -1;
    Int16   returnCode;
    Boolean haveR;
    Boolean first;
    UInt32  timeLimit;

    if ( controlSocket < 0 )
        return MY_ERROR;

    if ( NetUReadN( controlSocket, code, 3 ) < 3 )
        return MY_ERROR;

    returnCode = ( code[ 0 ] - '0' ) * 100 + ( code[ 1 ] - '0' ) * 10 +
                 ( code[ 2 ] - '0' );

//    code[ 3 ] = 0;

    scannedNumber = 3;
    haveR    = false;
    timeLimit     = TimGetTicks() + RESPONSE_TIMEOUT * SysTicksPerSecond();
    AppNetTimeout = RESPONSE_TIMEOUT * SysTicksPerSecond();
    lastLine      = false;
    first         = true;

    while ( TimGetTicks() < timeLimit ) {
        Int16 didRead;
        Int16 i;

        didRead = read( controlSocket, miscBuffer, MISC_BUFFER_SIZE - 1 );

        if ( didRead < 1 )
            return MY_ERROR;

        if ( first ) {
            UInt16 toCopy;

            toCopy = RESPONSE_START_SIZE - 1;

            if ( didRead < toCopy )
                toCopy = didRead;

            MemMove( responseStart, miscBuffer, toCopy );
            responseStart[ toCopy ] = 0;

            if ( *responseStart != 0 )
                Message( FROM_SERVER, responseStart + 1 );

            first = false;
        }

        miscBuffer[ didRead ] = 0;

        if ( haveR && miscBuffer[ 0 ] == '\n' ) {
            if ( lastLine )
                return returnCode;
            if ( miscBuffer[ 4 ] == ' ' )
                lastLine = true;
        }

        haveR = false;

        if ( 0 <= scannedNumber ) {
            i = 4 - scannedNumber;
            if ( miscBuffer[ i - 1 ] == ' ' ) {
                lastLine = true;
            }
        }
        else {
            i = 0;
        }

        scannedNumber = -1;

        for ( i = 0 ; i < didRead - 1 ; i++ ) {
             if ( miscBuffer[ i ] == '\r' && miscBuffer[ i + 1 ] == '\n' ) {
                 if ( lastLine ) {
                     return returnCode;
                 }
                 i += 2;
                 if ( i + 3 < didRead ) {
                     if ( miscBuffer[ i + 3 ] == ' ' ) {
                         lastLine = true;
                     }
                 }
                 else {
                     scannedNumber = didRead - i;
                 }
             }
        }
        
        if ( miscBuffer[ didRead - 1 ] == '\r' )
            haveR = true;
    }

    return MY_ERROR;
}



UInt16 FTPCommand( Char* command )
{
    AppNetTimeout = SEND_COMMAND_TIMEOUT * SysTicksPerSecond();

    if ( ! StrNCompare( command, "PASS ", 5 ) )
        Message( TO_SERVER, "PASS [...]" );
    else
        Message( TO_SERVER, command );

    if ( controlSocket < 0 ||
         NetUWriteN( controlSocket, command, StrLen( command ) ) < StrLen( command ) ) {
        Message( ERROR, "Error sending command." );
        return MY_ERROR;
    }

    return FTPResponse();
}


Boolean FTPPWD( Char* dest, UInt16 maxLength )
{
    Char* p;

    if ( 257 != FTPCommand( "PWD\r\n" ) ) {
        return false;
    }

    if ( ! *responseStart )
        return false;

    p = responseStart + 1;

    if ( *p != '"' )
        return false;

    p++;

    while ( maxLength-- && *p ) {
        if ( *p == '"' && p[ 1 ] == '"' )
            p++;
        else if ( *p == '"' )
            break;
        *dest++ = *p++;
    }

    if ( *p != '"' ) {
        return false;
    }
    else {
        *dest = 0;
        return true;
    }

}



Boolean FTPPassive( void )
{
    Char* s;
    UInt8 data[ 6 ];
    UInt16 i;

    if ( 227 != FTPCommand( "PASV\r\n" ) )
        return false;

    s = StrChr( miscBuffer, '(' );

    if ( s != NULL )
        s++;

    for ( i = 0 ; i < 6 && s != NULL && *s ; i++ ) {
         data[ i ] = StrAToI( s );

         while ( '0' <= *s && *s <= '9' )
             s++;

         while ( *s && ! ( '0' <= *s && *s <= '9' ) )
             s++;
    }

    if ( i < 6 )
        return false;

    passiveAddress.family = AF_INET;

    passiveAddress.port = data[ 4 ] * 256 + data[ 5 ];
    passiveAddress.addr = 0;
    for ( i = 0 ; i < 4 ; i++ )
         passiveAddress.addr = passiveAddress.addr * 256 + data[ i ];

    havePassive = true;

    return true;
}



static void ClosePassiveDataSocket( void )
{
    if ( 0 <= dataSocket ) {
        close( dataSocket );
        dataSocket = -1;
    }
}



static Boolean OpenPassiveDataSocket( void )
{
    ClosePassiveDataSocket();

    if ( /* ! havePassive && */ ! FTPPassive() )
        return false;

    dataSocket = socket( AF_INET, netSocketTypeStream, 0 );;

    if ( dataSocket < 0 )
        return false;

    Linger( dataSocket, true, 0 );

    if ( errNone != connect( dataSocket, &passiveAddress, sizeof( passiveAddress ) ) ) {
        ClosePassiveDataSocket();
        return false;
    }

    return true;
}



Boolean FTPBinary( Boolean binary )
{
    return FTPCommand( binary ? "TYPE I\r\n" : "TYPE A\r\n" ) < 400;
}



Boolean FTPRename( const Char* src, const Char* dest )
{
    StrPrintF( command, "RNFR %s\r\n", src );

    if ( FTPCommand( command ) < 400 )
        return false;

    StrPrintF( command, "RNTO %s\r\n", src );

    return FTPCommand( command ) < 400;
}



static Err MyVFSFileRead( UInt32 ref, UInt32 size, void* buf, UInt32* didReadP )
{
    return VFSFileRead( ( FileRef )ref, size, buf, didReadP );
}



static Err MyVFSFileWrite( UInt32 ref, UInt32 size, const void* buf, UInt32* didWriteP )
{
    return VFSFileWrite( ( FileRef )ref, size, buf, didWriteP );
}



static Err BufferWrite( UInt32 ref, UInt32 size, const void* buf, UInt32* didWriteP )
{
    if ( writeBufferSize == 1 ) {
        *writeBuffer = 0;
        writeBuffer++;
        writeBufferSize--;
    }
    else if ( 1 < writeBufferSize ) {
        UInt32 toWrite;

        toWrite = size;

        if ( writeBufferSize - 1 < size ) {
            toWrite = writeBufferSize - 1;
        }

        if ( 0 < toWrite ) {
            MemMove( writeBuffer, buf, toWrite );
            writeBuffer[ toWrite ] = 0;
            writeBuffer += toWrite;
            writeBufferSize -= toWrite;
        }
    }
    *didWriteP = size;
    return errNone;
}



static Boolean FTPSend( const Char* dest, ReaderFunction Reader, UInt32 ref, UInt32 size )
{
    Err     err;
    Boolean success;

    if ( FTP_COMMAND_LENGTH < StrLen( dest ) + 1 + 4 + 2 + 1 )
        return false;

    success = false;

    if ( OpenPassiveDataSocket() ) {
        UInt8* buffer;

        buffer = MemPtrNew( DATA_BUFFER_SIZE );

        if ( buffer != NULL ) {
            UInt32 readSoFar;
            
//            if ( 36 <= StrLen( dest ) )
//                StrCat( dest, "..t" );

            StrPrintF( command, "STOR %s\r\n", dest );

            if ( 400 <= FTPCommand( command ) )
                return false;

            if ( size == 0 )
                size = 0x7FFFFFFFu;

            readSoFar = 0;

            do {
                UInt32 didRead;

                ProgressBar( 1, readSoFar, size );
                err = Reader( ref, DATA_BUFFER_SIZE, buffer, &didRead );

                if ( ( err == errNone || err == vfsErrFileEOF ) && 0 < didRead ) {
                    if ( didRead != NetUWriteN( dataSocket, buffer, didRead ) ) {
                        err = 1;
                        break;
                    }
                }

                readSoFar += didRead;

            } while ( err == errNone );

            ProgressBar( 1, 1, 1 );

            if ( err != errNone && err != vfsErrFileEOF ) {
                FTPCommand( "ABOR\r\n" );
                StrPrintF( command, "DELE %s\r\n", dest );
                FTPCommand( command );
                StrPrintF( miscBuffer, "(Error %x)", err );
                Message( ERROR, miscBuffer );
                SysTaskDelay( SysTicksPerSecond() );
            }
            else {
                success = true;
            }

            MemPtrFree( buffer );
        }

        if ( shutdown( dataSocket, SHUT_RDWR ) < 0 ) {
            Message( ERROR, "Error shutting down connection." );
            SysTaskDelay( SysTicksPerSecond() );
            success = false;
        }

        if ( success ) {
            success = FTPResponse() < 300;
        }

        ClosePassiveDataSocket();
    }

    return success;
}



Boolean FTPSendFromVFS( const Char* dest, UInt16 vol, const Char* path )
{
    FileRef ref;
    Boolean success;
    UInt32  size;

    if ( errNone != VFSFileOpen( vol, path, vfsModeRead, &ref ) )
        return false;

    if ( errNone != VFSFileSize( ref, &size ) )
        size = 0;

    success = FTPSend( dest, MyVFSFileRead, ref, size );

    VFSFileClose( ref );

    return success;
}


/* Does not use command[] */
static Boolean FTPFetchData( Char* fetchCommand, WriterFunction Writer, UInt32 ref )
{
    Err     err;
    Boolean success;

    success = false;

    if ( OpenPassiveDataSocket() ) {
        UInt8* buffer;

        buffer = MemPtrNew( DATA_BUFFER_SIZE );

        if ( buffer != NULL ) {
            Int32 didRead = 0;

            if ( 400 <= FTPCommand( fetchCommand ) )
                return false;

            do {
                UInt32 didWrite;

                didRead = NetUReadN( dataSocket, buffer, DATA_BUFFER_SIZE );

                if ( didRead < 0 ) {
                    if ( errno == 0 )
                        err = 1;
                    else
                        err = errno;
                    break;
                }
                else {
                    err = errNone;
                }

                if ( err == errNone && 0 < didRead ) {
                    err = Writer( ref, didRead, buffer, &didWrite );
                }

            } while ( err == errNone && didRead == DATA_BUFFER_SIZE );

            if ( err != errNone ) {
                FTPCommand( "ABOR\r\n" );
            }
            else {
                success = true;
            }

            MemPtrFree( buffer );
        }

        if ( shutdown( dataSocket, SHUT_RDWR ) < 0 )
            success = false;

        if ( success ) {
            success = FTPResponse() < 300;
        }

        ClosePassiveDataSocket();
    }

    return success;
}



Boolean FTPFetchToVFS( Char* remoteName, UInt16 vol, const Char* localPath )
{
    FileRef ref;
    Boolean success;

    if ( FTP_COMMAND_LENGTH <= StrLen( remoteName ) + 8 )
        return false;

    VFSFileCreate( vol, localPath );

    if ( errNone != VFSFileOpen( vol, localPath, vfsModeWrite, &ref ) )
        return false;

    StrPrintF( command, "RETR %s\r\n", remoteName );

    success = FTPFetchData( command, MyVFSFileWrite, ref );

    VFSFileClose( ref );

    return success;
}



// Does not handle filenames containing \r.
static Err DirSave( UInt32 ref, UInt32 toSave, const void* buffer, UInt32* didSaveP )
{
    const Char*  p;
    UInt32 i;

    if ( curName == NULL ) {
        curNamePos = 0;
        curName = MemPtrNew( MAX_FILENAME + 1 );
        if ( curName == NULL ) {
            return 1;
        }
        *curName = 0;
    }

    p = buffer;

    for ( i = 0 ; i < toSave ; i++ ) {
         if ( p[ i ] == '\r' ) {
             curName[ curNamePos ] = 0;

             if ( dirListStart == NULL ) {
                 dirListStart = MemPtrNew( sizeof( *dirListStart ) );
                 dirListCur   = dirListStart;
             }

             if ( dirListCur == NULL ) {
                 return 1;
             }
             else {
                 /* dirListCur always points beyond the list */
                 dirListCur->next = MemPtrNew( sizeof( *dirListCur ) );
                 if ( dirListCur->next == NULL )
                     return 1;
                 MemPtrResize( curName, curNamePos + 1 );
                 dirListCur->name = curName;
                 dirListCur = dirListCur->next;
                 dirListCur->name = NULL;
                 dirListCur->next = NULL;
                 curName = MemPtrNew( MAX_FILENAME + 1 );
                 if ( curName == NULL ) {
                     return 1;
                 }
                 curNamePos = 0;
                 *curName = 0;
             }
         }
         else if ( p[ i ] != '\n' ) {
             if ( curNamePos < MAX_FILENAME ) {
                 curName[ curNamePos++ ] = p[ i ];
             }
             else {
                 return 1;
             }
         }
    }

    return errNone;
}



static void ClearDirList( void )
{
    if ( curName != NULL ) {
        MemPtrFree( curName );
        curName = NULL;
        curNamePos = 0;
    }

    while ( dirListStart != NULL ) {
         DirListNode* next;

         next = dirListStart->next;

         SafeMemPtrFree( dirListStart->name );
         SafeMemPtrFree( dirListStart );

         dirListStart = next;
    }

    dirListCur = NULL;
}



Boolean FTPFetchString( const Char* name, UInt32 bufferSize, Char* s )
{
    if ( FTP_COMMAND_LENGTH <= StrLen( name ) + 8 )
        return false;

    StrPrintF( command, "RETR %s\r\n", name );

    writeBuffer     = s;
    writeBufferSize = bufferSize;

    if ( 0 < bufferSize )
        *s = 0;

    return FTPFetchData( command, BufferWrite, 0 );
}



Char** FTPDirList( Int32* numEntriesP, Boolean subDirOnly )
{
    UInt32 count;
    UInt32 i;
    UInt32 j;
    DirListNode* n;
    Char**       names;

    ClearDirList();

    *numEntriesP = 0;

    if ( ! FTPFetchData( "NLST\r\n", DirSave, 0 ) || dirListStart == NULL ) {
        return NULL;
    }

    count = 0;

    n = dirListStart;

    while ( n != NULL && n->name != NULL && n->name[ 0 ] != 0 ) {
        count++;
        n = n->next;
    }

    if ( count == 0 ) {
        ClearDirList();

        return NULL;
    }

    names = MemPtrNew( count * sizeof( Char* ) );

    if ( names == NULL ) {
        ClearDirList();

        return NULL;
    }

    n = dirListStart;
    j = 0;

    for ( i = 0 ; i < count && n != NULL ; i++ ) {
         DirListNode* n1;
         
         if ( n->name != NULL ) {
             if ( subDirOnly ) {
                 if ( StrCompare( n->name, "." ) &&
                      StrCompare( n->name, ".." ) &&
                      FTPChDir( n->name ) ) {
                                              
                     names[ j++ ] = n->name;
                     FTPChDir( ".." ); // error handling!
                 }
             }
             else {
                 names[ j++ ] = n->name;
             }
         }

         n1         = n->next;
         MemPtrFree( n );
         n          = n1;
         if ( n->name == NULL ) {
             break;
         }

    }

    count = j;

    dirListStart = NULL;
    dirListCur   = NULL;
    
    SafeMemPtrFree( curName );
    curName = NULL;
    curNamePos = 0;

    *numEntriesP = count;

    return names;
}



Boolean FTPMkDir( Char* dir )
{
    StrPrintF( command, "MKD %s\r\n", dir );

    return FTPCommand( command ) < 400;
}



Boolean FTPDelete( Char* file )
{
    StrPrintF( command, "DELE %s\r\n", file );

    return FTPCommand( command ) < 400;
}



Boolean FTPRmDir( Char* dir )
{
    StrPrintF( command, "RMD %s\r\n", dir );

    return FTPCommand( command ) < 400;
}



Boolean FTPChDir( Char* dir )
{
    if ( dir == NULL || ! *dir )
        return true;

    if ( ! StrCompare( dir, ".." ) ) {
        return FTPCommand( "CDUP\r\n" ) < 400;
    }
    else {
        StrPrintF( command, "CWD %s\r\n", dir );

        return FTPCommand( command ) < 400;
    }
}



void Linger( NetSocketRef socket, Boolean state, UInt16 time )
{
    struct linger l;

    l.l_onoff = state;
    l.l_linger  = time;

    setsockopt( socket, SOL_SOCKET, SO_LINGER, &l, sizeof( l ) );
}




Boolean NetOpen( void )
{
//    UInt32 ifCreator;
//    UInt16 ifInstance;
    Err    err;
    UInt16 returnValue;

    NetClose();

    errno = 0;

    err = SysLibFind( "Net.lib", &AppNetRefnum );

    if ( err != errNone ) {
        return false;
    }

#if 0
    err = NetLibIFGet( AppNetRefnum, 0, &ifCreator, &ifInstance );

    if ( err != errNone ) {
        return false;
    }
#endif    

    err = NetLibOpen( AppNetRefnum, &returnValue );

    if ( err != netErrAlreadyOpen && err != errNone ) {
        return false;
    }

    oldTimeout = AppNetTimeout;

    netOpen    = true;

    return true;
}



Boolean FTPOpen( Char* hostName, Char* userName, Char* password, UInt16 port )
{
    Int16 returnValue;
    extern Boolean locked;

    if ( ! netOpen )
        return false;

    locked = PrefGetPreference( prefDeviceLocked );

    FTPClose();

    if ( port == 0 )
         controlSocket = NetUTCPOpen( hostName, "ftp", 0 );
    else
         controlSocket = NetUTCPOpen( hostName, NULL, port );

    if ( controlSocket < 0 ) {
 //       NetLibClose( AppNetRefnum, false );
        return false;
    }

    Linger( controlSocket, false /* true */, 0 );

    returnValue = FTPResponse();

    if ( returnValue < 0 || 400 <= returnValue ) {
        FTPClose();
        return false;
    }

    StrPrintF( command, "USER %s\r\n", userName );
    returnValue = FTPCommand( command );

    if ( returnValue < 0 || returnValue == MY_ERROR ) {
        FTPClose();
        return false;
    }

    if ( 400 <= returnValue ) {
        FTPClose();
        return false;
    }

    if ( 200 <= returnValue ) {
        Char myPassword[ MAX_PASSWORD_SIZE + 1 ];
        Char* p;

        if ( password == NULL ) {
            FrmCustomResponseAlert( altPasswordFTP, "", "", "",
                myPassword, MAX_PASSWORD_SIZE, NULL );
            p = myPassword;
        }
        else {
            p = password;
        }

        StrPrintF( command, "PASS %s\r\n", p );
        returnValue = FTPCommand( command );

        if ( 300 <= returnValue ) {
            FTPClose();
            return false;
        }
    }

    FTPBinary( true );

    return true;
}
