/* SysZLib.c -- handle opening and closing of the SysZLib shared library,
 * and branching into various versions of the ZLib functions
 *
 *  Original version by Tom Zerucha (tz@execpc.com) 2000
 *  Revisions and ARMlet support by Justin Clark (jclark@copera.com) 2003
 */

#include <PalmOS.h>
#include <PalmCompatibility.h>
#include "PceNativeCall.h"
#include "SysZLib.h"
#include "zlib_armlet_got.h"
#include "zlib_armlet.h"
#include "endianutils.h"


/* Location of the SysZLib x86 DLL for the Simulator to use */
#define SYSZLIB_DLL_NAME "zlib_armlet.dll\0zlib_armlet_dispatch"


/* Structure for storing all the global data that SysZLib needs */
typedef struct SysZLibGlobalsType_s
{
  long got[sizeof(zlib_armlet_got) / sizeof(long)]; /* Adjusted GOT */  
  zlib_armlet_params armlet_params; /* Struct for passing params to ARMlet */
  unsigned short refNum;   /* Shared library reference number */
  short openCount;         /* Number of times the library has been opened */
  MemHandle armletH;       /* Handle of the ARM code resource */
  void * armletP;          /* Locked pointer to the ARM code resource */
  Boolean useSimulatorDLL; /* Flag indicating we should call the x86 DLL */

} SysZLibGlobalsType;


/* ===========================================================================
 * Set up globals variables if necessary, detect execution environment, and
 * keep track of how many times library has been opened.
 */
Err zlib_open (UInt16 ref) 
{
  UInt32 processorID;
  SysZLibGlobalsType * gP;
  UInt32 i;
  long * old_got = (long *)zlib_armlet_got;
  UInt32 paceVersion;
  DmOpenRef dbref;

  /* Check if globals exist already. */
  gP = SysLibTblEntry(ref)->globalsP;
  if (gP != NULL)
  {
    gP->openCount++;
    return 0;
  }

  /* Create new globals. */
  gP = MemPtrNew (sizeof(SysZLibGlobalsType));
  if (gP == NULL)
    return SysZLibErrOutOfMemory;

  MemPtrSetOwner (gP, 0);
  gP->openCount = 1;
  gP->refNum = ref;
  gP->armletH = 0;
  gP->armletP = 0;
  gP->useSimulatorDLL = 0;
  SysLibTblEntry(ref)->globalsP = gP;

  /* Check for ARM or x86 processor. */
  if (!FtrGet(sysFtrCreator, sysFtrNumProcessorID, &processorID))
  {
#ifdef USE_DLL
    if (processorID == sysFtrNumProcessorx86)
    {
      /* We're on x86 */
      gP->useSimulatorDLL = 1;
    }
    else
#endif
    if (sysFtrNumProcessorIsARM(processorID))
    {
      /* Check for PACE. */
      if (!FtrGet('pace', 0, &paceVersion))
      {
	/* Find and lock armlet resource. */
	dbref = DmOpenDatabaseByTypeCreator ('libr', 'ZLib', dmModeReadOnly);
	gP->armletH = DmGetResource ('ARMC', 1000);
	if (gP->armletH)
	{
	  gP->armletP = MemHandleLock (gP->armletH);

	  /* Fix the global offset table. */
	  for (i = 0; i < sizeof(zlib_armlet_got) / sizeof(long); i++)
	  {
	    gP->got[i] = (long)ByteSwap32(old_got[i]) + (long)gP->armletP;
	    gP->got[i] = ByteSwap32(gP->got[i]);
	  }

	  gP->armlet_params.gotP = gP->got;
	}
	DmCloseDatabase (dbref);
      }
    }
  }

  return 0;
}


/* ===========================================================================
 * Decrement open count, clean up globals if count reaches zero, return
 * new open count.
 */
Err zlib_close (UInt16 ref, UInt16 *numop) 
{
  SysZLibGlobalsType * gP = SysLibTblEntry(ref)->globalsP;
  if (gP == NULL)
    return SysZLibErrNotOpen;

  gP->openCount--;
  *numop = gP->openCount;

  if (gP->openCount == 0)
  {
    /* Cleanup */
    if (gP->armletP) MemPtrUnlock (gP->armletP);
    if (gP->armletH) DmReleaseResource (gP->armletH);
    MemPtrFree (gP);
    SysLibTblEntry(ref)->globalsP = NULL;
  }

  return errNone;
}


/* ========================================================================= */
Err zlib_null (UInt16 ref) 
{
  return 0;
}


/* ===========================================================================
 * Prototypes for the 68K versions of the ZLib functions
 */

#undef deflate
#undef deflateEnd
#undef inflate
#undef inflateEnd

int deflateInit2_ (unsigned short ref, z_streamp strm, int level, 
		   int method, int windowBits, int memLevel, int strategy, 
		   const char * version, int stream_size);
int deflate (unsigned short ref, z_streamp strm, int flush);
int deflateEnd (unsigned short ref, z_streamp strm);
int inflateInit2_ (unsigned short ref, z_streamp z, int w, 
		   const char * version, int stream_size);
int inflate (unsigned short d, z_streamp z, int f);
int inflateEnd (unsigned short d, z_streamp z);

/* Hack to get globals and make sure zlib_open was actually called. This is 
   to make up for the bad ZLSetup macro that some people used, which wasn't
   actually calling zlib_open. */

#define GET_LIBRARY_GLOBALS \
  SysZLibGlobalsType * gP = SysLibTblEntry(ref)->globalsP; \
  if (gP == NULL) \
  { \
    zlib_open (ref); \
    gP = SysLibTblEntry(ref)->globalsP; \
  }


/* ===========================================================================
 * The following functions get called by user apps and branch into the
 * appropriate version of the actual ZLib functions.
 */
int deflateInit2_SZ (unsigned short ref, z_streamp strm, int level, 
		     int method, int windowBits, int memLevel, int strategy, 
		     const char * version, int stream_size)
{
  GET_LIBRARY_GLOBALS
  
  if (gP->useSimulatorDLL || gP->armletP)
  {
    gP->armlet_params.function = ZLAF_DEFLATEINIT2;
    gP->armlet_params.strm = strm;
    gP->armlet_params.level = level;
    gP->armlet_params.method = method;
    gP->armlet_params.windowBits = windowBits;
    gP->armlet_params.memLevel = memLevel;
    gP->armlet_params.strategy = strategy;
    gP->armlet_params.version = (char *)version;

    if (gP->useSimulatorDLL) 
    {
      return PceNativeCall ((NativeFuncType *)SYSZLIB_DLL_NAME, 
			    &(gP->armlet_params));
    }
    else 
    {
      return PceNativeCall (gP->armletP, &(gP->armlet_params));
    }
  }

  return deflateInit2_ (ref, strm, level, method, windowBits, memLevel, 
			strategy, version, stream_size);
}


/* ========================================================================= */
int deflateSZ (unsigned short ref, z_streamp strm, int flush)
{
  GET_LIBRARY_GLOBALS

  if (gP->useSimulatorDLL || gP->armletP)
  {
    gP->armlet_params.function = ZLAF_DEFLATE;
    gP->armlet_params.strm = strm;
    gP->armlet_params.flush = flush;
    
    if (gP->useSimulatorDLL) 
    {
      return PceNativeCall ((NativeFuncType *)SYSZLIB_DLL_NAME, 
			    &(gP->armlet_params));
    }
    else 
    {
      return PceNativeCall (gP->armletP, &(gP->armlet_params));
    }
  }

  return deflate (ref, strm, flush);
}


/* ========================================================================= */
int deflateEndSZ (unsigned short ref, z_streamp strm)
{
  GET_LIBRARY_GLOBALS

  if (gP->useSimulatorDLL || gP->armletP)
  {
    gP->armlet_params.function = ZLAF_DEFLATEEND;
    gP->armlet_params.strm = strm;
      
    if (gP->useSimulatorDLL) 
    {
      return PceNativeCall ((NativeFuncType *)SYSZLIB_DLL_NAME, 
			    &(gP->armlet_params));
    }
    else 
    {
      return PceNativeCall (gP->armletP, &(gP->armlet_params));
    }
  }
  
  return deflateEnd (ref, strm);
}


/* ========================================================================= */
int inflateInit2_SZ (unsigned short ref, z_streamp strm, int windowBits, 
		     const char * version, int stream_size)
{
  GET_LIBRARY_GLOBALS

  if (gP->useSimulatorDLL || gP->armletP)
  {
    gP->armlet_params.function = ZLAF_INFLATEINIT2;
    gP->armlet_params.strm = strm;
    gP->armlet_params.windowBits = windowBits;
    gP->armlet_params.version = (char *)version;
    
    if (gP->useSimulatorDLL) 
    {
      return PceNativeCall ((NativeFuncType *)SYSZLIB_DLL_NAME, 
			    &(gP->armlet_params));
    }
    else
    {
      return PceNativeCall (gP->armletP, &(gP->armlet_params));
    }
  }

  return inflateInit2_ (ref, strm, windowBits, version, stream_size);
}


/* ========================================================================= */
int inflateSZ (unsigned short ref, z_streamp strm, int flush)
{
  GET_LIBRARY_GLOBALS

  if (gP->useSimulatorDLL || gP->armletP)
  {
    gP->armlet_params.function = ZLAF_INFLATE;
    gP->armlet_params.strm = strm;
    gP->armlet_params.flush = flush;
    
    if (gP->useSimulatorDLL) 
    {
      return PceNativeCall ((NativeFuncType *)SYSZLIB_DLL_NAME, 
			    &(gP->armlet_params));
    }
    else 
    {
      return PceNativeCall (gP->armletP, &(gP->armlet_params));
    }
  }

  return inflate (ref, strm, flush);
}


/* ========================================================================= */
int inflateEndSZ (unsigned short ref, z_streamp strm)
{
  GET_LIBRARY_GLOBALS

  if (gP->useSimulatorDLL || gP->armletP)
  {
    gP->armlet_params.function = ZLAF_INFLATEEND;
    gP->armlet_params.strm = strm;
      
    if (gP->useSimulatorDLL) 
    {
      return PceNativeCall ((NativeFuncType *)SYSZLIB_DLL_NAME, 
			    &(gP->armlet_params));
    }
    else 
    {
      return PceNativeCall (gP->armletP, &(gP->armlet_params));
    }
  }
  
  return inflateEnd (ref, strm);
}

